# Print stuff from elements.xml
#
# Suggestions and improvements are solicited, and should be sent to michael
# at the domain name of landcroft dot co dot uk.
#
# Currently, the command line is
#
#	$ ruby printelts.rb [-d directory] [-l n] [-c "Category name"] [-t]
#		[-p pdfile] [-h htmlfile] [-f font_for_HTML] [-m n] [-r n] -s -n -x -i
#
# in which the options are:
#	-d	your FullRecall directory
#	-l	print n most difficult (hardest) items
#	-r	print items more difficult than n
#	-c	print all items of the given category
#	-t	uses tabs rather than spaces to set columns in simple output
#	-m	use difficulty metric n (n in [0..0])
#	-s	show category column
#	-f	font to use for HTML output
#	-n	report on the numbers of items shown and total
#	-x	extended formatted statistics, always to stderr
#	-i	output id numbers only
#
# If you specify -p or -h, the default (simple) output is not produced.
#
# PDF is limited to 8-bit characters by the PDF::Writer library.
#
# One point on which user-input is sought is the difficulty algorithm in sub
# difficultness --- other users might have further suggestions.
#
# Another matter is horizontal alignment in the plain text output;  currently,
# the script presumes that the font is mono-spaced, which is clearly wrong, but
# I know of no other way of calculating column-alignment.  Option -t helps here.
#
# Currently, the script just outputs questions and answers in UTF-8, formatted
# into columns.  There is the possibility of converting this into PDF but
# currently the software does not work (for me).  Instructions for doing this:
#
# 1.  You need a BDF font;  convert a TTF using http://crl.nmsu.edu/~mleisher/ttf2bdf.html
#
#	$ otf2bdf-LINUX-3.0/otf2bdf -r 800 -v ./.fonts/Code2000.ttf -o Code2000.bdf
#
# 2.  UTF-8 Text to PS convertor:  http://oldrus-ispell.sourceforge.net/txtbdf2ps.html, development version.
#
#	$ ruby THISCRIPT.rb | ./txtbdf2ps-dev.pl --bdf=Code2000.bdf -UTF-8 -page=A4 -tab=30 -nowrap >sample.ps
#
# The problem I experienced was that the txtbdf2ps-dev script outputs Postscript
# that GhostScript cannot convert.  GS 8.14 is alright;  8.57 is not.
#
# Versions:
# 1.0 2007.Aug.01 initial release
# 1.1 2007.Aug.08 added -t -p -h output options (old -h is now -l)
# 1.2 2007.Aug.09 fixed bug in tabbing on plain text.
# 1.3 2007.Aug.09 added options -s, -m and -r, and grade column
# 1.4 2007.Aug.11 fixed bug when using option -c
# 1.5 2007.Aug.14 added more options for difficulty calculation
# 1.6 2007.Aug.15 revise diff. calc. and allow specification of HTML font
# 1.7 2007.Aug.24 added new difficulty calculation (5) and -m option
# 1.8 2007.Aug.26 removed unused diff. calcs. and added statistics reports
# 1.9 2007.Sep.06 corrrected HTML UTF-8 declaration
#
# This Ruby implementation is to be superceded by the Haskell implementation
# available from the FR files area.  It is not intended to develop this Ruby
# version any more.

require "rexml/document"
require "rexml/streamlistener"

require 'time'
TimeStamp = Time.now.to_f


class Difficulty

	def initialize(&function)
		@function = function
	end

	def calculation(attributes)
		@function.call(attributes)
	end

end

def repetitions(attributes)
	attributes["rp"].to_f
end

def lifetime(attributes)
	TimeStamp - attributes["id"].to_f
end

def metric_luohan(attributes)
	Math.sqrt(interval_ratio(attributes) * repeats_ratio(attributes))
end

def interval_ratio(attributes)
	interval = attributes["ivl"].to_f
	repeats = repetitions(attributes)
	repeats * repeats / (interval + 1)
end

def repeats_ratio(attributes)
	repetitions(attributes) / ((lifetime(attributes) / 86400) ** 0.38)
end

# Currently we only have one metric ...
Difficultness = [
	Difficulty.new {|attributes|
		Math.log(1 + metric_luohan(attributes))
	}
]

CatName = "cat"
GradeName = "gr"

Col_heads = { CatName => "Category", "Q" => "Question", "A" => "Answer", GradeName => "Grade" }


def de_8(text)
	text.scan(/./u).map{|utf8char|
		if utf8char.length == 1 then
			utf8char
		elsif utf8char.length == 2 then
			value = (utf8char[0] % 32) * 64 + (utf8char[1] % 64)
			if value < 0x100 then
				value.chr
			else
				'?'
			end
		else
			'?'
		end
	}.join
end

def de_8_pdf(text)
	de_8(text).gsub(/<[bus]>/, "{").gsub(/<\/[bus]>/, "}")
end

class Listener

	include REXML::StreamListener

	def tag_start(name, attributes)
		@tag_stack << name
		case name
		when "fullrecall"
			@version = attributes["core_version"].to_i
			if @version != 10 and @version != 12 then
				warn "Unknown version;  results not guaranteed"
			end
		when "q", "a"
			# for simplicity, we always use upper-case Q and A
			@tag_stack[-1].upcase!
		when "Q", "A"
			# nothing
		when "category"
			if @version == 12
				@cat_stack << attributes["name"]
			end
		when "item"
			# analyse attributes to determine WoN this item be `difficult'
			@items += 1
			@item_attrs = attributes
		else
			warn "Unexpected tag <" + name + ">"
		end
	end

	def tag_end(name)
		@tag_stack.pop
		case name
		when "item"
			@item_part[CatName] =
				case @version
				when 10
					@item_attrs["ctgr"]
				when 12
					@cat_stack.join(" / ")
				end
			@item_part["id"] = @item_attrs["id"]
			@item_part[GradeName] = $schwierigkeit.calculation(@item_attrs)
			@diff_sum += @item_part[GradeName]
			item
			@item_part = { }
		when "fullrecall"
			if $columns.key?(GradeName) then
				@results.each {|item|
					image = (item[GradeName] * 100).round.to_s
					item[GradeName] = (image.length < 3) ?  '0.' + '0'*(2 - image.length) + image : image.insert(-3, '.')
				}
			end
			if $pee_dee_eff != nil then
				print_PDF
			end
			if $aitch_tee_em_ell != nil then
				print_HTML
			end
			if $print_just_ids then
				print_ids_only
			end
			if $simple_output then
				print_plain_text
			end
			if $extended_stats then
				$stderr.print Time.now.to_f, ";" ,@results.length, ";", @items, ";", sprintf("%.3f", 100.0 * @results.length / @items), "%;", @diff_sum / @items, "\n"
			end
		when "category"
			if @version == 12
				@cat_stack.pop
			end
		end
	end

	def text(wording)
		if @tag_stack[-1] =~ /^[QA]$/ then
			@item_part[@tag_stack[-1]] = wording
		elsif wording.gsub(/[ \n\r]/, "").length > 0 then
			warn "unexpected text \"" + wording + "\" in <" + @tag_stack[-1] + ">"
		end
	end

	def cdata(wording)
		text(wording.sub(/<img>.*<\/img>/, ""))
	end

	def print_plain_text
		$stdout.print "FullRecall items\n"
		@results.each{|result|
			$stdout.print '=' * $all_width, "\n"
			more = true
			# In this loop we print all lines of an item.  An item will
			# have multiple lines in either Q or A part if the number
			# of characters exceed the column-width, or if the item
			# contain a new-line character.
			itempart = { }
			ColHeads.map{|what|
				itempart[what] = result[what].gsub(/<[bus]>/, "{").gsub(/<\/[bus]>/, "}")
			}
			while more do
				$tabbing.reset
				# Does this item have more lines to print ?
				more = false
				ColHeads.each{|what|
					# Slice out the maximum number of UTF-8 chars.,
					# and test for a new-line.
					# The pattern //u is UTF-8 matching.
					this_line = itempart[what].scan(/.{0,#{$columns[what]}}/mu)[0]
					trailing = this_line.sub!(/\n.*$/mu, "") != nil
					length = this_line.scan(/./u).size
					if this_line.length > 0 then
						$stdout.print $tabbing.to_s, this_line
						$tabbing.reset
					end
					# inter-column space
					$tabbing.increment($columns[what] - length)
					# Slice off what we just printed, plus any new-line
					itempart[what].slice!(0..this_line.length-(trailing ? 0 : 1))
					# If anything remain, there is more to print
					more ||= itempart[what].length > 0
				}
				$stdout.print "\n"
			end
		}
		if $show_numbers then
			$stdout.print "\nItems: ", @results.length, "/", @items, " (", 100.0 * @results.length / @items, "%)\n"
		end
	end

	def print_ids_only
		@results.each{|result|
			$stdout.print result["id"], "\n"
		}
	end

	def print_PDF
		require "rubygems"
		require "pdf/writer"
		require "pdf/simpletable"

		pdf = PDF::Writer.new
		pdf.select_font("Helvetica")

		PDF::SimpleTable.new do |tab|
			tab.title = "FullRecall Items"
			tab.column_order = ColHeads

			$columns.each_key {|key|
				tab.columns[key] = PDF::SimpleTable::Column.new(key) { |col|
					col.heading = Col_heads[key]
				}
			}

			tab.show_lines    = :all
			tab.show_headings = true
			tab.orientation   = :center
			tab.position      = :center

			data = @results.map{|item|
				line = { "Q" => de_8_pdf(item["Q"]), "A" => de_8_pdf(item["A"]) }
				if $columns.key?(CatName) then
					line[CatName] = item[CatName]
				end
				if $columns.key?(GradeName) then
					line[GradeName] = item[GradeName]
				end
				line
			}

			tab.data.replace data
			tab.render_on(pdf)
		end
		if $show_numbers then
			pdf.text ""
			pdf.text "Items: " + @results.length.to_s + "/" + @items.to_s + " (" + (100.0 * @results.length / @items).to_s + "%)\n"
		end
		pdf.save_as($pee_dee_eff)
	end

	def print_HTML
		html = File.new($aitch_tee_em_ell, "w")
		html.print "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
		html.print "<html><font face=\"", $font_HTML, "\">\n<head>\n"
		html.print "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"/>\n"
		html.print "<title>FullRecall Items</title></head>\n"
		html.print "<table width=\"100%\" cellpadding=\"2\" cellspacing=\"0\" border=\"1\">\n"
		html.print "<tr bgcolor=\"#ffcc00\">", ColHeads.map{|key| "<th>"+ Col_heads[key] + "</th>"}, "</tr>\n"
		@results.each{|result|
			html.print "<tr>", (ColHeads.map{|what| "<td>" + result[what].gsub(/\n/, "<br/>") + "</td>"}), "</tr>\n"
		}
		html.print "</table>\n"
		if $show_numbers then
			html.print "<br>Items: ", @results.length, "/", @items, " (", 100.0 * @results.length / @items, "%)\n"
		end
		html.print "</body>\n</html>\n"
	end

	def initialize(*argument)
		@cat_stack = [ ]
		@results = [ ]
		@tag_stack = [ ]
		@item_part = { }
		@parameter = argument[0]
		@items = 0
		@diff_sum = 0
	end
end


class Difficult < Listener

	def item
		difficulty = @item_part[GradeName]
		insertion = 0
		new_minimum = true
		while new_minimum and insertion < @results.length do
			if difficulty > @results[insertion][GradeName] then
				@results.insert(insertion, @item_part)
				if @results.length > @parameter then
					@results.delete_at(@parameter)
				end
				new_minimum = false
			end
			insertion += 1
		end
		if new_minimum then
			if @results.length < @parameter then
				@results << @item_part
			end
		end
	end

end


class HarderThan < Listener

	def item
		difficulty = @item_part[GradeName]
		if difficulty > @parameter then
			if @results.length == 0 or difficulty < @results[-1][GradeName] then
				@results << @item_part
			else
				ctr = 0
				while @results[ctr][GradeName] > difficulty do
					ctr += 1
				end
				@results.insert(ctr, @item_part)
			end
		end
	end
end


class Category < Listener

	def item
		if @item_part[CatName] == @parameter then
			@results << @item_part
		end
	end
end


class Inter_column_white

	def initialize
		@quantity = 0
	end

	def to_s
		@character * @quantity
	end

	def reset
		@quantity = 0
	end
end


class Inter_space < Inter_column_white

	def initialize
		@character = ' '
		super
	end

	def increment(howmuch)
		@quantity += 1 + howmuch
	end

end


class Inter_tab < Inter_column_white

	def initialize
		@character = "\t"
		super
	end

	def increment(howmuch)
		@quantity += 1
	end

end

$font_HTML = "Code2000"
$schwierigkeit = Difficultness[0]
$tabbing = Inter_space.new
$pee_dee_eff = nil
$aitch_tee_em_ell = nil
$simple_output = true
directory = ""
collector = Difficult.new(50)
$columns = { "Q" => 35, "A" => 48 }
$show_numbers = false
$extended_stats = false
$print_just_ids = false

while ARGV.length > 0 do
	case ARGV.shift
	when '-t'
		$tabbing = Inter_tab.new
	when '-f'
		$font_HTML = ARGV.shift
	when '-r'
		collector = HarderThan.new(ARGV.shift.to_f)
		$columns[GradeName] = 7
	when '-l'
		collector = Difficult.new(ARGV.shift.to_i)
	when '-c'
		collector = Category.new(ARGV.shift)
	when '-s'
		$columns[CatName] = 20
	when '-n'
		$show_numbers = true
	when '-x'
		$extended_stats = true
	when '-m'
		$schwierigkeit = Difficultness[ARGV.shift.to_i]
	when '-h'
		$simple_output = false
		base = ARGV.shift
		$aitch_tee_em_ell = (base =~ /\.html$/ ? base : base + '.html')
	when '-i'
		$simple_output = false
		$print_just_ids = true
	when '-p'
		$simple_output = false
		base = ARGV.shift
		$pee_dee_eff = (base =~ /\.pdf$/ ? base : base + '.pdf')
	when '-d'
		base = ARGV.shift
		directory = (base =~ /[\/\\]$/ ? base : base + '/')
	end
end

# Set ColHeads to those columns that will appear in the output
ColHeads = [ CatName, "Q", "A", GradeName].select {|colhead| $columns.key?(colhead) }

$all_width = 0
$columns.each_value {|wide| $all_width += wide }

REXML::Parsers::StreamParser.new(File.new(directory + "elements.xml", "r"), collector).parse
